#!/usr/bin/env python3
import sys
import os
import docopt
import logging
import termcolor
from . import __doc__ as pkg_doc, __name__ as root_name, DocoptError, __version__
from .parser import ParserParameters, Parser
from .script import Script

log = logging.getLogger(root_name)

__doc__ = pkg_doc + """
Usage:
  docopt.sh [options] (- | SCRIPT)
  docopt.sh generate-library

Options:
  --line-length -n N   Max line length when minifying (0 to disable,default: 80)
  --library -l SRC     Generates only the dynamic part of the parser and
                       includes the static parts using `source SRC`.
                       Use `generate-library` to create that file.
  --no-auto-params -P  Disable auto-detection parser generation parameters
  --parser -p          Output the parser instead of inserting it in the script
  --help -h            Show this help screen
  --version            Show the docopt.sh version

Note:
  When reading the script from stdin (`-` instead of SCRIPT)
  docopt.sh will output the modified script to stdout.

Parameters:
  You can set the global variables before invoking docopt with
  `eval "$(docopt "$@")"` to change the behavior of the parser.
  Consult the readme for advanced options.

  $DOCOPT_PROGRAM_VERSION  The string to print when --version is specified
                           [default: none/disabled]
  $DOCOPT_ADD_HELP         Set to `false` to not print usage on --help
                           [default: true]
  $DOCOPT_OPTIONS_FIRST    Set to `true` to fail when options are specified
                           after arguments/commands [default: false]
"""


def docopt_sh(params):
  # `generate-library` will never be true because it is specified after [SCRIPT]
  # which matches it first. We want it in that order though, but there's a simple workaround:
  if params['SCRIPT'] == 'generate-library':
    parser_parameters = ParserParameters(params)
    parser = Parser(parser_parameters)
    sys.stdout.write('#!/usr/bin/env bash\n\n' + str(parser.generate_library(add_version_check=True)))
  else:
    try:
      if params['-']:
        script = Script(sys.stdin.read())
      else:
        with open(params['SCRIPT'], 'r') as h:
          script = Script(h.read(), params['SCRIPT'])
      script.validate()

      parser_parameters = ParserParameters(params, script)
      parser = Parser(parser_parameters)

      if params['--parser']:
        sys.stdout.write(parser.generate(script))
      else:
        patched_script = script.patch(parser)
        if params['-']:
          sys.stdout.write(str(patched_script))
        else:
          with open(params['SCRIPT'], 'w') as h:
            h.write(str(patched_script))
          if patched_script == script:
            log.info('The parser in %s is already up-to-date.', params['SCRIPT'])
          else:
            log.info('%s has been updated.', params['SCRIPT'])
    except DocoptError as e:
      log.error(str(e))
      sys.exit(e.exit_code)


def setup_logging():
  level_colors = {
    logging.ERROR: 'red',
    logging.WARN: 'yellow',
  }

  class ColorFormatter(logging.Formatter):

    def format(self, record):
        record.msg = termcolor.colored(record.msg, level_colors.get(record.levelno, None))
        return super(ColorFormatter, self).format(record)

  stderr = logging.StreamHandler(sys.stderr)
  if os.isatty(2):
    stderr.setFormatter(ColorFormatter())
  log.setLevel(level=logging.INFO)
  log.addHandler(stderr)


def main():
  setup_logging()
  params = docopt.docopt(__doc__, version=__version__)
  docopt_sh(params)

if __name__ == '__main__':
  main()



import subprocess
import shlex
import os
import io
from tempfile import NamedTemporaryFile
from collections import namedtuple
from contextlib import contextmanager
from docopt_sh.__main__ import main as docopt_sh_main
from docopt_sh.bash import bash_variable_value

Usecase = namedtuple('Usecase', ['line', 'bash', 'doc', 'prog', 'argv', 'expect'])


def bash_eval_script(bash, script, argv):
  argv = ' '.join(map(shlex.quote, argv))
  process = subprocess.run(
    [bash[1], '-c', 'set - %s; eval "$(cat)"' % argv],
    input=script.encode('utf-8'),
    stdout=subprocess.PIPE, stderr=subprocess.PIPE,
    timeout=2
  )
  return process.returncode, process.stdout.decode('utf-8'), process.stderr.decode('utf-8')


def declare_quote(value):
  return value.replace('\\', '\\\\').replace('"', '\\"')


def replace_docopt_params(stream, docopt_params):
  script = stream.read()
  params = '\n'.join([k + '=' + bash_variable_value(v) for k, v in docopt_params.items()])
  script = script.replace('"DOCOPT PARAMS"', params)
  return io.StringIO(script)


def patch_stream(monkeypatch, capsys, stream, program_params=[], docopt_params={}):
  stream = replace_docopt_params(stream, docopt_params)
  captured = invoke_docopt(monkeypatch, capsys, stdin=stream, program_params=program_params + ['-'])

  def run(bash, *argv):
    return bash_eval_script(bash, captured.out, argv)
  return run


def patch_file(monkeypatch, capsys, name, program_params=[], docopt_params={}):
  with open(os.path.join('tests/scripts', name)) as handle:
    return patch_stream(monkeypatch, capsys, handle, program_params, docopt_params)


@contextmanager
def temp_file(name, docopt_params={}):
  with open(os.path.join('tests/scripts', name)) as handle:
    script = replace_docopt_params(handle, docopt_params).read()
  file = NamedTemporaryFile(mode='w', delete=False)
  try:
    file.write(script)
    file.close()

    def run(bash, *args):
      process = subprocess.run(
        [bash[1], file.name] + list(args),
        stdout=subprocess.PIPE, stderr=subprocess.PIPE
      )
      return process.returncode, process.stdout.decode('utf-8'), process.stderr.decode('utf-8')
    yield file, run
  finally:
    os.unlink(file.name)


@contextmanager
def generated_library(monkeypatch, capsys, program_params=[]):
  file = NamedTemporaryFile(mode='w', delete=False)
  try:
    captured = invoke_docopt(monkeypatch, capsys, program_params=['generate-library'] + program_params)
    file.write(captured.out)
    file.close()
    yield file
  finally:
    os.unlink(file.name)


def invoke_docopt(monkeypatch, capsys=None, program_params=[], stdin=None):
  with monkeypatch.context() as m:
    if stdin is not None:
      m.setattr('sys.stdin', stdin)
    m.setattr('sys.argv', ['docopt.sh'] + program_params)
    docopt_sh_main()
    if capsys is not None:
      return capsys.readouterr()
    return None



import logging
import os
import subprocess
import re
import glob
import json
from . import Usecase

log = logging.getLogger(__name__)


def pytest_addoption(parser):
  parser.addoption(
    '--bash-version', type=str,
    help='test with specific bash version (comma separated, or "all" for all installed versions)'
  )


def pytest_sessionstart(session):
  bash_version = session.config.getoption('--bash-version')
  if bash_version == 'all':
    session.config.bash_versions = get_installed_bash_versions()
    all_versions = [version for version, _ in session.config.bash_versions]
    log.warning('Testing with bash versions: %s' % ', '.join(all_versions))
  elif bash_version:
    session.config.bash_versions = [get_bash_version(version) for version in bash_version.split(',')]
    log.warning('Testing with bash versions: %s' % ', '.join(bash_version.split(',')))
  else:
    version, executable = get_system_bash()
    log.warning('Testing with system bash: %s' % version)
    session.config.bash_versions = [(version, executable)]


def pytest_generate_tests(metafunc):
  if 'bash' in metafunc.fixturenames:
    if metafunc.config.bash_versions:
      metafunc.parametrize('bash', metafunc.config.bash_versions)
  if 'usecase' in metafunc.fixturenames:
    with open(os.path.join(os.path.dirname(__file__), 'usecases.txt'), 'r') as handle:
      contents = handle.read()
    metafunc.parametrize('usecase', [usecase for usecase in parse_usecases(contents)])


def pytest_assertrepr_compare(config, op, left, right):
  if isinstance(left, Usecase) and isinstance(right, Usecase) and op == '==':
    error = ['Usecase on line %d failed' % left.line]
    error.append('bash: %s' % left.bash)
    error.append('%s' % left.doc)
    error.append('$ %s %s' % (left.prog, left.argv))
    if isinstance(left.expect, str) or isinstance(right.expect, str):
      error.append('%s != %s' % (repr(left.expect), repr(right.expect)))
    else:
      for key, value in left.expect.items():
        if key in right.expect:
          if value != right.expect[key]:
            error.append('%s != %s' % (repr(value), repr(right.expect[key])))
        else:
          error.append('%s != (key %s not found)' % (repr(value), key))
      for key, value in right.expect.items():
        if key not in left.expect:
          error.append('(key %s not found) != %s' % (key, repr(value)))
    return error


def parse_usecases(raw):
  fixture_pattern = re.compile(r'r"""(?P<doc>[^"]+)""".*?(?=r""")', re.DOTALL)
  case_pattern = re.compile(
    r'\$ (?P<prog>[^\n ]+)( (?P<argv>[^\n]+))?\n(?P<expect>[^\n]+?)(?P<comment>\s*#[^\n]*)?\n\n')
  for fixture_match in fixture_pattern.finditer(raw):
    line_offset = raw[:fixture_match.start(0)].count('\n') + 1
    fixture = fixture_match.group(0)
    doc = fixture_match.group('doc')
    for case_match in case_pattern.finditer(fixture):
      line = line_offset + fixture[:case_match.start(0)].count('\n')
      case = case_match.group(0)
      prog = case_match.group('prog')
      argv = case_match.group('argv')
      argv = argv if argv else ''
      try:
        expect = json.loads(case_match.group('expect'))
      except json.decoder.JSONDecodeError as e:
        json_line = line_offset + fixture[:case_match.start('expect')].count('\n')
        raise Exception('Error on line %d:\n%s\n---\n%s' % (json_line, case, str(e)))
      yield Usecase(line, None, doc, prog, argv, expect)


def get_system_bash():
  process = subprocess.run(
    ['bash', '--version'],
    stdout=subprocess.PIPE, stderr=subprocess.PIPE
  )
  process.check_returncode()
  out = process.stdout.decode('utf-8')
  match = re.search(r'GNU bash, version ([^ ]+)', out)
  if not match:
    log.warning('Unable to detect installed bash version from:\n%s\nAssuming 4.4' % out)
    return '4.4', 'bash'
  else:
    return match.group(1), 'bash'


def get_installed_bash_versions():
  versions = [get_system_bash()]
  for executable in glob.glob('tests/bash-versions/bash-*/bash'):
    folder_name = os.path.basename(os.path.dirname(executable))
    versions.append((re.search(r'bash-(.+)', folder_name).group(1), executable))
  return versions


def get_bash_version(version):
  executable_path = os.path.join(os.path.dirname(__file__), 'bash-versions/bash-%s/bash' % version)
  if not os.path.exists(executable_path):
    raise Exception(
      'bash-{version} is not installed. '
      'Use `tests/get_bash.py {version}` to do that'
      .format(version=version)
    )
  return version, executable_path



import os.path
import re
import hashlib
import logging
import shlex
from collections import OrderedDict
from . import __version__, DocoptError
from .doc_ast import DocAst, Option
from .bash import Code, indent, bash_ifs_value, minify
from .node import LeafNode

log = logging.getLogger(__name__)


class Parser(object):

  def __init__(self, parser_parameters):
    self.parameters = parser_parameters
    self.library = Library()
    self.shellcheck_ignores = [
      '2016',  # Ignore unexpanded variables in single quotes (used for docopt_exit generation)
    ]
    if self.parameters.library_path:
      self.shellcheck_ignores.extend([
        '1091',  # Ignore library sourcing
        '2034',  # Ignore unused vars (they refer to things in the library)
      ])

  def generate(self, script):
    generated = self.generate_main(script)
    if not self.parameters.library_path:
      generated = generated + self.generate_library()
    if self.parameters.minify:
      generated = generated.minify(self.parameters.max_line_length)
    return str(generated)

  def generate_main(self, script):
    if self.parameters.library_path:
      library_source = 'source %s \'%s\'' % (self.parameters.library_path, __version__)
    else:
      library_source = ''

    doc_value_start, doc_value_end = script.doc.stripped_value_boundaries
    stripped_doc = '${{DOC:{start}:{length}}}'.format(
      start=doc_value_start,
      length=doc_value_end - doc_value_start,
    )

    doc_ast = DocAst(script.doc.value)
    usage_start, usage_end = doc_ast.usage_match
    usage_doc = '${{DOC:{start}:{length}}}'.format(
      start=str(doc_value_start + usage_start),
      length=str(usage_end - usage_start),
    )

    leaf_nodes = [n for n in doc_ast.nodes if type(n) is LeafNode]
    option_nodes = [node for node in leaf_nodes if type(node.pattern) is Option]

    replacements = {
      '"LIBRARY SOURCE"': library_source,
      '    "OUTPUT TEARDOWN"\n': '' if library_source else "    printf 'docopt_do_teardown\\n'\n",
      '"DOC VALUE"': stripped_doc,
      '"DOC USAGE"': usage_doc,
      '"DOC DIGEST"': hashlib.sha256(script.doc.value.encode('utf-8')).hexdigest()[0:5],
      '"SHORTS"': ' '.join([bash_ifs_value(o.pattern.short) for o in option_nodes]),
      '"LONGS"': ' '.join([bash_ifs_value(o.pattern.long) for o in option_nodes]),
      '"ARGCOUNT"': ' '.join([bash_ifs_value(o.pattern.argcount) for o in option_nodes]),
      '    "NODES"': indent('\n'.join(map(str, list(doc_ast.nodes))), level=2),
      '    "DEFAULTS"': indent('\n'.join([node.default_assignment for node in leaf_nodes]), level=2),
      '"VAR NAMES"': ' \\\n    '.join(['"%s"' % node.prefixed_variable_name for node in leaf_nodes]),
      '"HAS VARS"': bash_ifs_value(True if leaf_nodes else False),
      '"MAX NODE IDX"': max([n.idx for n in doc_ast.nodes]),
      '"ROOT NODE IDX"': doc_ast.root_node.idx,
    }
    return self.library.main.replace_literal(replacements)

  def generate_library(self, add_version_check=False):
    functions = self.library.functions
    if add_version_check:
      functions['lib_version_check'] = functions['lib_version_check'].replace_literal(
        {'"LIBRARY VERSION"': __version__})
    else:
      del functions['lib_version_check']
    return Code(functions.values())


class Library(object):

  def __init__(self):
    function_re = re.compile((
        r'^(?P<name>[a-z_][a-z0-9_]*)\(\)\s*\{'
        r'\n+'
        r'(?P<body>.*?)'
        r'\n+\}\n$'
      ), re.MULTILINE | re.IGNORECASE | re.DOTALL)
    self.functions = OrderedDict([])
    with open(os.path.join(os.path.dirname(__file__), 'docopt.sh'), 'r') as handle:
      for match in function_re.finditer(handle.read()):
        name = match.group('name')
        if name == 'docopt':
          self.main = Code(match.group(0))
        elif name == 'lib_version_check':
          self.functions[name] = Code(match.group('body'))
        else:
          self.functions[name] = Code(match.group(0))


class ParserParameter(object):

  def __init__(self, name, invocation_params, script_params, default):
    if script_params is None:
      script_value = None
    else:
      script_value = script_params[name]
    defined_in_invocation = invocation_params[name] is not None
    defined_in_script = script_value is not None
    auto_params = not invocation_params['--no-auto-params']

    self.name = name
    self.defined = (defined_in_invocation or defined_in_script) and auto_params
    self.invocation_value = invocation_params[name] if defined_in_invocation else default
    self.script_value = script_value if defined_in_script else default
    self.merged_from_script = not defined_in_invocation and auto_params and defined_in_script
    self.value = self.script_value if self.merged_from_script else self.invocation_value
    self.changed = script_params is not None and self.value != self.script_value

  def __str__(self):
    return '%s=%s' % (self.name, shlex.quote(self.value))


class ParserParameters(object):

  def __init__(self, invocation_params, script=None):
    if script is not None:
      script_params = script.guards.bottom.refresh_command_params
      if script_params is None:
        if script.guards.present and not invocation_params['--no-auto-params']:
          raise DocoptError(
            'Unable to auto-detect parser generation parameters. '
            'Re-run docopt.sh with `--no-auto-params`.'
          )
    else:
      script_params = None

    params = OrderedDict([])
    params['--line-length'] = ParserParameter('--line-length', invocation_params, script_params, default='80')
    params['--library'] = ParserParameter('--library', invocation_params, script_params, default=None)

    merged_from_script = list(filter(lambda p: p.merged_from_script, params.values()))
    if merged_from_script:
      log.info(
        'Adding `%s` from parser generation parameters that were detected in the script. '
        'Use --no-auto-params to disable this behavior.',
        ' '.join(map(str, merged_from_script))
      )

    self.max_line_length = int(params['--line-length'].value)
    self.library_path = params['--library'].value
    self.minify = self.max_line_length > 0

    command = ['docopt.sh']
    command_short = ['docopt.sh']
    if params['--line-length'].defined:
      command.append(str(params['--line-length']))
    if params['--library'].defined:
      command.append(str(params['--library']))
    if script is not None and script.path:
      command.append(os.path.basename(script.path))
      command_short.append(os.path.basename(script.path))
    else:
      command.append('-')
      command.append('<FILE')
      command_short.append('-')
      command_short.append('<FILE')
    self.refresh_command = ' '.join(command)
    self.refresh_command_short = ' '.join(command_short)



import re
import logging
import shlex
import docopt
from . import DocoptError

log = logging.getLogger(__name__)


class Script(object):

  def __init__(self, contents, path=None):
    self.contents = contents
    self.path = path

    self.doc = Doc(self)
    self.guards = Guards(self, self.doc)
    self.invocation = Invocation(self, self.guards)
    self.options = [Option(self, name) for name in [
      'DOCOPT_ADD_HELP', 'DOCOPT_PROGRAM_VERSION', 'DOCOPT_OPTIONS_FIRST',
      'DOCOPT_PREFIX', 'DOCOPT_TEARDOWN', 'DOCOPT_DOC_CHECK', 'DOCOPT_LIB_CHECK'
    ]]

  def validate(self):
    if not self.doc.present:
      raise DocoptScriptValidationError(
        'Could not find variable containing usage doc. '
        'Make sure your script contains a `DOC="... Usage: ..."` variable',
        self.doc
      )
    if self.doc.count > 1:
      raise DocoptScriptValidationError(
        'More than one variable containing usage doc found. '
        'Search your script for `DOC=`, there should be only one such declaration.',
        self.doc
      )
    guard_help = (
        'Search your script for `# docopt parser below/above`, '
        'there should be exactly one with `below` and one with `above` (in that order). '
        'If in doubt, just delete anything that is not your code and try again.'
    )
    if self.guards.top.count > 1:
      raise DocoptScriptValidationError(
        'Multiple docopt parser top guards found. ' + guard_help, self.guards.top)
    if self.guards.bottom.count > 1:
      raise DocoptScriptValidationError(
        'Multiple docopt parser bottom guards found.' + guard_help, self.guards.bottom)
    if self.guards.top.present and not self.guards.bottom.present:
        raise DocoptScriptValidationError(
          'Parser top guard found, but no bottom guard detected. ' + guard_help, self.guards.top)
    if self.guards.bottom.present and not self.guards.top.present:
      raise DocoptScriptValidationError(
        'Parser bottom guard found, but no top guard detected. ' + guard_help, self.guards.bottom)
    if self.invocation.count > 1:
      log.warning(
        '%s Multiple invocations of docopt found, check your script to make sure this is correct.',
        self.invocation
      )
    if not self.invocation.present:
      log.warning(
        '%s No invocations of docopt found, check your script to make sure this is correct.\n'
        'docopt.sh is invoked with `eval "$(docopt "$@")"`.',
        self.invocation
      )
    for option in self.options:
      if self.invocation.present and option.present and option.start > self.invocation.last.end:
        log.warning(
          '%s $%s has no effect when specified after invoking docopt, '
          'make sure to place docopt options before calling `eval "$(docopt "$@")"`.',
          option, option.name
        )

  def patch(self, parser):
    return Script(
      "{start}{guard_begin}\n{shellcheck_ignores}\n{parser}{guard_end}\n{end}".format(
        start=self.contents[:self.guards.start],
        guard_begin=(
          "# docopt parser below, refresh this parser with `%s`"
          % parser.parameters.refresh_command_short),
        shellcheck_ignores='# shellcheck disable=%s' % ','.join(parser.shellcheck_ignores),
        parser=parser.generate(self),
        guard_end=(
          "# docopt parser above, complete command for generating this parser is `%s`"
          % parser.parameters.refresh_command),
        end=self.contents[self.guards.end:],
      )
    )

  def __eq__(self, other):
    return self.contents == other.contents

  def __str__(self):
    return self.contents


class ScriptLocation(object):

  def __init__(self, script, matches, offset):
    self.script = script
    self.matches = list(matches)
    self.match = self.matches[0] if self.matches else None
    self.offset = 0 if offset is None else offset
    self.present = self.match is not None
    self.count = len(self.matches)
    self.start = self.match.start(0) + self.offset if self.present else None
    self.end = self.match.end(0) + self.offset if self.present else None
    self.line = self.script.contents[:self.start].count('\n') + 1
    self.all = [self] + [ScriptLocation(self.script, iter([match]), self.offset) for match in self.matches[1:]]
    if self.count == 0:
      self.last = None
    elif self.count == 1:
      self.last = self
    else:
      self.last = ScriptLocation(self.script, iter([self.matches[-1]]), self.offset)

  def __len__(self):
    return self.end - self.start if self.present else 0

  def __str__(self):
    if not self.present:
      return '%s' % self.script.path
    if self.count > 1:
      return '%s:%s' % (self.script.path, ','.join(map(lambda l: str(l.line), self.all)))
    else:
      return '%s:%d' % (self.script.path, self.line)


class Doc(ScriptLocation):

  def __init__(self, script):
    matches = re.finditer(
      r'DOC="(\s*)(.*?Usage:.+?)(\s*)"(\n|;)',
      script.contents,
      re.MULTILINE | re.IGNORECASE | re.DOTALL
    )
    # re.IGNORECASE causes doc=, Doc= etc. to be matched, remove those matches
    matches = filter(lambda m: m.group(0).startswith('DOC='), matches)
    super(Doc, self).__init__(script, matches, 0)
    self.value = self.match.group(2) if self.present else None
    self.stripped_value_boundaries = (
      self.match.start(2) - self.match.start(1),
      self.match.end(2) - self.match.start(1)
    ) if self.present else None


class TopGuard(ScriptLocation):

  def __init__(self, script, doc):
    matches = re.finditer(
      r'# docopt parser below(, refresh this parser with `([^`]+)`)?.*\n',
      script.contents[doc.end:],
      re.MULTILINE
    )
    super(TopGuard, self).__init__(script, matches, doc.end)


class BottomGuard(ScriptLocation):

  def __init__(self, script, top):
    matches = re.finditer(
      r'# docopt parser above(, complete command for generating this parser is `([^`]+)`)?.*\n',
      script.contents[top.end:],
      re.MULTILINE
    )
    super(BottomGuard, self).__init__(script, matches, top.end)
    self.refresh_command_params = None
    if self.present and self.match.group(2) is not None:
      from .__main__ import __doc__
      try:
        self.refresh_command_params = docopt.docopt(__doc__, shlex.split(self.match.group(2))[1:])
      except (docopt.DocoptLanguageError, docopt.DocoptExit) as e:
        pass


class Guards(object):

  def __init__(self, script, doc):
    self.top = TopGuard(script, doc)
    if self.top.present:
      self.bottom = BottomGuard(script, self.top)
    else:
      self.bottom = BottomGuard(script, doc)
    self.present = self.top.present and self.bottom.present
    # The top.offset is to easily handle the absence of a parser
    self.start = self.top.start if self.present else self.top.offset
    self.end = self.bottom.end if self.present else self.top.offset

  def __len__(self):
    return self.end - self.start if self.present else 0


class Invocation(ScriptLocation):

  def __init__(self, script, parser):
    matches = re.finditer(r'eval "\$\(docopt\s+"\$\@"\)"', script.contents[parser.end:])
    super(Invocation, self).__init__(script, matches, parser.end)


class Option(ScriptLocation):

  def __init__(self, script, name):
    self.name = name
    matches = re.finditer(r'^%s=' % name, script.contents, re.MULTILINE)
    super(Option, self).__init__(script, matches, 0)
    if self.count > 1:
      # Override parent class selection of first match, previous assignments
      # would be overwritten so it's the last match that has an effect
      self.match = self.matches[-1]


class DocoptScriptValidationError(DocoptError):

  def __init__(self, message, script_location=None):
    # Exit code 74: input/output error (sysexits.h)
    super(DocoptScriptValidationError, self).__init__(message, exit_code=74)
    self.script_location = script_location

  def __str__(self):
    if self.script_location is not None:
      return '%s %s' % (self.script_location, self.message)
    return self.message



from .doc_ast import Option, Command, Required, Optional, OptionsShortcut, OneOrMore, Either
from .bash import Code, bash_variable_name, bash_variable_value, bash_ifs_value

helper_map = {
  Required: 'docopt_required',
  Optional: 'docopt_optional',
  OptionsShortcut: 'docopt_optional',
  OneOrMore: 'docopt_oneormore',
  Either: 'docopt_either',
}


class Node(Code):

  def __init__(self, pattern, body, idx):
    self.pattern = pattern
    self.idx = idx
    code = '{name}(){{\n{body}\n}}\n'.format(
      name='docopt_node_' + str(idx),
      body=body,
    )
    super(Node, self).__init__(code)


class BranchNode(Node):

  def __init__(self, pattern, idx, node_map):
    # minify arg list by only specifying node idx
    child_indexes = map(lambda child: node_map[child].idx, pattern.children)
    body = '  {helper} {args}'.format(
      helper=helper_map[type(pattern)],
      args=' '.join(list(map(str, child_indexes))),
    )
    super(BranchNode, self).__init__(pattern, body, idx)


class LeafNode(Node):

  def __init__(self, pattern, idx):
    default_value = pattern.value
    if type(pattern) is Option:
      helper_name = 'docopt_switch' if type(default_value) in [bool, int] else 'docopt_value'
      needle = idx
    elif type(pattern) is Command:
      helper_name = 'docopt_command'
      needle = pattern.name
    else:
      helper_name = 'docopt_value'
      needle = 'a'
    self.variable_name = bash_variable_name(pattern.name)

    args = [self.variable_name, bash_ifs_value(needle)]
    if type(default_value) in [list, int]:
      args.append(bash_ifs_value(True))
    elif helper_name == 'docopt_command' and args[0] == args[1]:
      args = [args[0]]
    body = '  {helper} {args}'.format(
      helper=helper_name,
      args=' '.join(args),
    )

    if type(default_value) is list:
      default_tpl = (
        'if declare -p {docopt_name} >/dev/null 2>&1; then\n'
        '  eval "${{docopt_prefix}}"\'{name}=("${{{docopt_name}[@]}}")\'\n'
        'else\n'
        '  eval "${{docopt_prefix}}"\'{name}={default}\'\n'
        'fi'
      )
    else:
      default_tpl = (
        'eval "${{docopt_prefix}}"\'{name}=${{{docopt_name}:-{default}}}\''
      )
    self.default_assignment = default_tpl.format(
      name=self.variable_name,
      docopt_name='docopt_var_' + self.variable_name,
      default=bash_variable_value(default_value)
    )
    self.prefixed_variable_name = '${{docopt_prefix}}{name}'.format(name=self.variable_name)
    super(LeafNode, self).__init__(pattern, body, idx)



import re
import shlex
import io
from docopt_sh.bash import bash_variable_name
from . import Usecase, patch_stream


def test_usecase(monkeypatch, capsys, usecase, bash):
  assert convert_to_bash(bash, usecase) == run_usecase(monkeypatch, capsys, usecase, bash)


def run_usecase(monkeypatch, capsys, usecase, bash):
  lineno, _, doc, prog, argv, expect = usecase
  program_template = '''
DOC="{doc}"
"DOCOPT PARAMS"
eval "$(docopt "$@")"
if [[ -n ${{!_usecase_*}} ]]; then
  declare -p "${{!_usecase_@}}"
fi
'''
  program = program_template.format(doc=doc)
  run = patch_stream(
    monkeypatch, capsys,
    io.StringIO(program),
    docopt_params={'DOCOPT_PREFIX': '_usecase_'}
  )
  code, out, err = run(bash, *shlex.split(argv))
  if code == 0:
    expr = re.compile('^declare (--|-a) ([^=]+)=')
    out = out.strip('\n')
    result = {}
    if out != '':
      for line in out.split('\n'):
        if expr.match(line) is None:
          raise Exception('Unable to match %s for usecase on line %d' % (line, lineno))
        result[expr.match(line).group(2)] = line
  else:
    result = 'user-error'
  return Usecase(lineno, bash[0], doc, prog, argv, result)


def convert_to_bash(bash, usecase):
  lineno, _, doc, prog, argv, expect = usecase
  if expect == 'user-error':
    declarations = expect
  else:
    declarations = {}
    for key, value in expect.items():
      var = '_usecase_' + re.sub(r'^[^a-z_]|[^a-z0-9_]', '_', key, 0, re.IGNORECASE)
      declarations[var] = bash_decl(bash[0], var, value)
  return Usecase(lineno, bash[0], doc, prog, argv, declarations)


def bash_decl(bash_version, name, value):
  if value is None or type(value) in (bool, int, str):
    return 'declare -- {name}={value}'.format(name=name, value=bash_decl_value(bash_version, value))
  if type(value) is list:
    return 'declare -a {name}={value}'.format(name=name, value=bash_decl_value(bash_version, value))
  raise Exception('Unknown value type %s' % type(value))


def bash_decl_value(bash_version, value):
  if value is None:
    return '""'
  if type(value) is bool:
    return '"true"' if value else '"false"'
  if type(value) is int:
    return '"{value}"'.format(value=value)
  if type(value) is str:
    return '"{value}"'.format(value=shlex.quote(value).strip("'"))
  if type(value) is list:
    list_tpl = '({value})'
    if bash_version is not None and (int(bash_version[0]) < 4 or int(bash_version[2]) < 4):
      list_tpl = "'({value})'"
    return list_tpl.format(value=' '.join('[{i}]={value}'.format(
      i=i, value=bash_decl_value(bash_version, v)) for i, v in enumerate(value))
    )



import re
import os
from io import StringIO
from . import bash_eval_script, patch_file, invoke_docopt, temp_file, generated_library
from docopt_sh.script import Script


def test_arg(monkeypatch, capsys, bash):
  run = patch_file(monkeypatch, capsys, 'echo_ship_name.sh')
  code, out, err = run(bash, 'ship', 'new', 'Britannica')
  assert code == 0
  assert out == 'Britannica\n'


def test_wrong_usage(monkeypatch, capsys, bash):
  run = patch_file(monkeypatch, capsys, 'echo_ship_name.sh')
  code, out, err = run(bash, '--bad-opt')
  assert code != 0


def test_help(monkeypatch, capsys, bash):
  run = patch_file(monkeypatch, capsys, 'echo_ship_name.sh')
  code, out, err = run(bash, '--help')
  assert code == 0
  assert out == 'Usage: echo_ship_name.sh ship new <name>...\n'
  assert err == ''


def test_no_help(monkeypatch, capsys, bash):
  run = patch_file(
    monkeypatch, capsys, 'echo_ship_name.sh',
    docopt_params={'DOCOPT_ADD_HELP': False}
  )
  code, out, err = run(bash, '--help')
  assert code == 1
  assert out == ''
  assert err == 'Usage: echo_ship_name.sh ship new <name>...\n'


def test_version(monkeypatch, capsys, bash):
  run = patch_file(
    monkeypatch, capsys, 'echo_ship_name.sh',
    docopt_params={'DOCOPT_PROGRAM_VERSION': '0.1.5'}
  )
  code, out, err = run(bash, '--version')
  assert code == 0
  assert out == '0.1.5\n'


def test_no_version(monkeypatch, capsys, bash):
  run = patch_file(
    monkeypatch, capsys, 'echo_ship_name.sh',
    docopt_params={'DOCOPT_PROGRAM_VERSION': False}
  )
  code, out, err = run(bash, '--version')
  assert code == 1
  assert out == ''
  assert err == 'Usage: echo_ship_name.sh ship new <name>...\n'


def test_options_anywhere(monkeypatch, capsys, bash):
  run = patch_file(monkeypatch, capsys, 'naval_fate.sh')
  code, out, err = run(bash, 'ship', 'Titanic', 'move', '1', '--speed', '6', '4')
  assert out == 'The Titanic is now moving to 1,4 at 6 knots.\n'


def test_options_first(monkeypatch, capsys, bash):
  run = patch_file(
    monkeypatch, capsys, 'naval_fate.sh',
    docopt_params={'DOCOPT_OPTIONS_FIRST': True}
  )
  code, out, err = run(bash, '--speed', '6', 'ship', 'Titanic', 'move', '1', '4')
  assert code == 0
  assert out == 'The Titanic is now moving to 1,4 at 6 knots.\n'


def test_options_first_fail(monkeypatch, capsys, bash):
  run = patch_file(
    monkeypatch, capsys, 'naval_fate.sh',
    docopt_params={'DOCOPT_OPTIONS_FIRST': True}
  )
  code, out, err = run(bash, 'ship', 'Titanic', 'move', '1', '--speed', '6', '4')
  assert code == 1
  assert out == ''
  assert err[:6] == 'Usage:'


def test_prefix(monkeypatch, capsys, bash):
  run = patch_file(
    monkeypatch, capsys, 'prefixed_echo.sh',
    docopt_params={'DOCOPT_PREFIX': 'prefix_', 'DOCOPT_TEARDOWN': False}
  )
  code, out, err = run(bash, 'ship', 'new', 'Titanic')
  assert code == 0
  assert out == 'Titanic\n'


def test_patch_file(monkeypatch, bash):
  with temp_file('echo_ship_name.sh') as (script, run):
    invoke_docopt(monkeypatch, program_params=[script.name])
    code, out, err = run(bash, 'ship', 'new', 'Olympia')
    assert out == 'Olympia\n'


def test_doc_check(monkeypatch, bash):
  with temp_file('echo_ship_name.sh') as (script, run):
    invoke_docopt(monkeypatch, program_params=[script.name])
    with open(script.name, 'r') as h:
      contents = h.read()
    contents = contents.replace('ship new <name>', 'ship delete <name>')
    with open(script.name, 'w') as h:
      h.write(contents)
    code, out, err = run(bash, 'ship', 'new', 'Olympia')
    regex = (
      r'^The current usage doc \([^)]+\) does not match what the parser was '
      r'generated with \([^)]+\)\nRun `docopt.sh` to refresh the parser.\n$'
    )
    assert re.match(regex, err) is not None


def test_no_doc_check(monkeypatch, bash):
  with temp_file('echo_ship_name.sh', docopt_params={'DOCOPT_DOC_CHECK': False}) as (script, run):
    invoke_docopt(monkeypatch, program_params=[script.name])
    with open(script.name, 'r') as h:
      contents = h.read()
    contents = contents.replace('ship new <name>', 'ship delete <name>')
    with open(script.name, 'w') as h:
      h.write(contents)
    code, out, err = run(bash, 'ship', 'new', 'Olympia')
    assert out == 'Olympia\n'


def test_parser_only(monkeypatch, capsys, bash):
  with open('tests/scripts/naval_fate.sh') as h:
    script = h.read()
  doc = Script(script).doc.value
  parser = invoke_docopt(monkeypatch, capsys=capsys, program_params=['--parser', '-'], stdin=StringIO(script)).out
  program = '''
DOC="{doc}"
{parser}
eval "$(docopt "$@")"
echo $((_x_ + _y_))
'''.format(doc=doc, parser=parser)
  captured = invoke_docopt(monkeypatch, capsys, program_params=['-'], stdin=StringIO(program))
  code, out, err = bash_eval_script(bash, captured.out, ['ship', 'shoot', '3', '1'])
  assert code == 0
  assert out == '4\n'


def test_teardown(monkeypatch, capsys, bash):
  run = patch_file(monkeypatch, capsys, 'all_vars.sh')
  code, out, err = run(bash, 'ship', 'new', 'Britannica')
  assert code == 0
  allowed_vars = []
  allowed_fns = ['docopt', 'docopt_exit']
  for line in out.strip().split('\n'):
    if '=' in line:
      name, val = line.split('=', 1)
      assert not name.lower().startswith('docopt') or name in allowed_vars
    elif '()' in line:
      name, rest = line.split(' ', 1)
      assert not name.lower().startswith('docopt') or name in allowed_fns


def test_library(monkeypatch, capsys, bash):
  with generated_library(monkeypatch, capsys) as library:
    run = patch_file(
      monkeypatch, capsys, 'echo_ship_name.sh',
      program_params=['--library', library.name]
    )
    code, out, err = run(bash, 'ship', 'new', 'Britannica')
    assert code == 0
    assert err == ''
    assert out == 'Britannica\n'


def test_library_teardown(monkeypatch, capsys, bash):
  with generated_library(monkeypatch, capsys) as library:
    run = patch_file(
      monkeypatch, capsys, 'all_vars.sh',
      program_params=['--library', library.name]
    )
    code, out, err = run(bash, 'ship', 'new', 'Britannica')
    allowed_vars = []
    allowed_fns = ['docopt', 'docopt_exit']
    for line in out.strip().split('\n'):
      if '=' in line:
        name, val = line.split('=', 1)
        assert not name.lower().startswith('docopt') or name in allowed_vars
      elif '()' in line:
        name, rest = line.split(' ', 1)
        assert not name.lower().startswith('docopt') or name in allowed_fns


def test_library_version(monkeypatch, capsys, bash):
  with generated_library(monkeypatch, capsys) as library:
    with temp_file('echo_ship_name.sh') as (script, run):
      invoke_docopt(monkeypatch, program_params=['--library', library.name, script.name])
      with open(script.name, 'r') as h:
        contents = h.read()
      contents = re.sub(r"source (\S+) '([^']+)'", r"source \1 '0.0.0'", contents)
      with open(script.name, 'w') as h:
        h.write(contents)
      code, out, err = run(bash, 'ship', 'new', 'Olympia')
      regex = (
        r'^The version of the included docopt library \([^)]+\) does not '
        r'match the version of the invoking docopt parser \(0\.0\.0\)\n$'
      )
      assert re.match(regex, err) is not None


def test_auto_params(monkeypatch, capsys, bash):
  with temp_file('naval_fate.sh') as (script, run):
    out = invoke_docopt(monkeypatch, capsys=capsys, program_params=[script.name, '--line-length', '20'])
    code, out, err = run(bash, 'ship', 'shoot', '3', '1')
    assert code == 0
    first_run = None
    with open(script.name, 'r') as handle:
      first_run = handle.read()
    invoke_docopt(monkeypatch, program_params=[script.name])
    code, out, err = run(bash, 'ship', 'shoot', '3', '1')
    assert code == 0
    with open(script.name, 'r') as handle:
      assert first_run == handle.read()


def test_error_fn(monkeypatch, capsys, bash):
  run = patch_file(monkeypatch, capsys, 'enum_check.sh')
  code, out, err = run(bash, '--color', 'sometimes')
  assert code == 1
  assert out == ''
  assert err == '--color must be auto, always, or never\nUsage: enum_check.sh [options]\n'



import re
from collections import OrderedDict
from itertools import chain


class DocAst(object):

  def __init__(self, doc):
    from .node import BranchNode, LeafNode
    doc = doc
    root, self.usage_match = parse_doc(doc)
    node_map = OrderedDict([])
    param_sort_order = [Option, Argument, Command]
    unique_params = list(OrderedDict.fromkeys(root.flat(*param_sort_order)))
    sorted_params = sorted(unique_params, key=lambda p: param_sort_order.index(type(p)))
    # Enumerate leaf nodes first so that their function index & index in the params array match
    for idx, param in enumerate(sorted_params):
      node_map[param] = LeafNode(param, idx)
    offset = len(node_map)
    for idx, pattern in enumerate(iter(root)):
      if isinstance(pattern, BranchPattern):
        node_map[pattern] = BranchNode(pattern, offset + idx, node_map)

    self.root_node = node_map[root]
    self.nodes = node_map.values()


def parse_doc(doc):
  usage_sections = parse_section('usage:', doc)
  if len(usage_sections) == 0:
    raise DocoptLanguageError('"usage:" (case-insensitive) not found.')
  if len(usage_sections) > 1:
    raise DocoptLanguageError('More than one "usage:" (case-insensitive).')
  usage, usage_match = usage_sections[0]
  # Trim newlines in usage_match
  match_fix = re.search(r'\A\n*(.+?)\n*\Z', usage, re.MULTILINE | re.DOTALL)
  usage_match = usage_match.start(0) + match_fix.start(0), usage_match.start(0) + match_fix.end(0)

  options = parse_defaults(doc)
  pattern = parse_pattern(formal_usage(usage), options)
  pattern_options = set(pattern.flat(Option))
  for options_shortcut in pattern.flat(OptionsShortcut):
    doc_options = parse_defaults(doc)
    options_shortcut.children = list(set(doc_options) - pattern_options)

  return pattern.fix(), usage_match


class DocoptLanguageError(Exception):

  """Error in construction of usage-message by developer."""


class Pattern(object):

  def __eq__(self, other):
    return repr(self) == repr(other)

  def __hash__(self):
    return hash(repr(self))

  def fix(self):
    self.fix_identities()
    self.fix_repeating_arguments()
    return self

  def fix_identities(self, uniq=None):
    """Make pattern-tree tips point to same object if they are equal."""
    if not hasattr(self, 'children'):
      return self
    uniq = list(set(self.flat())) if uniq is None else uniq
    for i, child in enumerate(self.children):
      if not hasattr(child, 'children'):
        assert child in uniq
        self.children[i] = uniq[uniq.index(child)]
      else:
        child.fix_identities(uniq)

  def fix_repeating_arguments(self):
    """Fix elements that should accumulate/increment values."""
    either = [list(child.children) for child in transform(self).children]
    for case in either:
      for e in [child for child in case if case.count(child) > 1]:
        if type(e) is Argument or type(e) is Option and e.argcount:
          if e.value is None:
            e.value = []
          elif type(e.value) is not list:
            e.value = e.value.split()
        if type(e) is Command or type(e) is Option and e.argcount == 0:
          e.value = 0
    return self


def transform(pattern):
  """Expand pattern into an (almost) equivalent one, but with single Either.

  Example: ((-a | -b) (-c | -d)) => (-a -c | -a -d | -b -c | -b -d)
  Quirks: [-a] => (-a), (-a...) => (-a -a)

  """
  result = []
  groups = [[pattern]]
  while groups:
    children = groups.pop(0)
    parents = [Required, Optional, OptionsShortcut, Either, OneOrMore]
    if any(t in map(type, children) for t in parents):
      child = [c for c in children if type(c) in parents][0]
      children.remove(child)
      if type(child) is Either:
        for c in child.children:
          groups.append([c] + children)
      elif type(child) is OneOrMore:
        groups.append(child.children * 2 + children)
      else:
        groups.append(child.children + children)
    else:
      result.append(children)
  return Either(*[Required(*e) for e in result])


class LeafPattern(Pattern):

  """Leaf/terminal node of a pattern tree."""

  def __init__(self, name, value=None):
    self.name, self.value = name, value

  def __repr__(self):
    return '%s(%r, %r)' % (self.__class__.__name__, self.name, self.value)

  def __iter__(self):
    yield self

  def flat(self, *types):
    return [self] if not types or type(self) in types else []


class BranchPattern(Pattern):

  """Branch/inner node of a pattern tree."""

  def __init__(self, *children):
    self.children = list(children)

  def __repr__(self):
    return '%s(%s)' % (self.__class__.__name__, ', '.join(repr(a) for a in self.children))

  def __iter__(self):
    for child in chain(*map(iter, self.children)):
      yield child
    yield self

  def flat(self, *types):
    if type(self) in types:
      return [self]
    return sum([child.flat(*types) for child in self.children], [])


class Argument(LeafPattern):

  @classmethod
  def parse(class_, source):
    name = re.findall('(<\S*?>)', source)[0]
    value = re.findall('\[default: (.*)\]', source, flags=re.I)
    return class_(name, value[0] if value else None)


class Command(Argument):

  def __init__(self, name, value=False):
    self.name, self.value = name, value


class Option(LeafPattern):

  def __init__(self, short=None, long=None, argcount=0, value=False):
    assert argcount in (0, 1)
    self.short, self.long, self.argcount = short, long, argcount
    self.value = None if value is False and argcount else value

  @classmethod
  def parse(class_, option_description):
    short, long, argcount, value = None, None, 0, False
    options, _, description = option_description.strip().partition('  ')
    options = options.replace(',', ' ').replace('=', ' ')
    for s in options.split():
      if s.startswith('--'):
        long = s
      elif s.startswith('-'):
        short = s
      else:
        argcount = 1
    if argcount:
      matched = re.findall('\[default: (.*)\]', description, flags=re.I)
      value = matched[0] if matched else None
    return class_(short, long, argcount, value)

  @property
  def name(self):
    return self.long or self.short

  def __repr__(self):
    return 'Option(%r, %r, %r, %r)' % (self.short, self.long, self.argcount, self.value)


class Required(BranchPattern):
  pass


class Optional(BranchPattern):
  pass


class OptionsShortcut(Optional):
  """Marker/placeholder for [options] shortcut."""


class OneOrMore(BranchPattern):
  pass


class Either(BranchPattern):
  pass


class Tokens(list):

  def __init__(self, source):
    self += source.split() if hasattr(source, 'split') else source

  @staticmethod
  def from_pattern(source):
    source = re.sub(r'([\[\]\(\)\|]|\.\.\.)', r' \1 ', source)
    source = [s for s in re.split('\s+|(\S*<.*?>)', source) if s]
    return Tokens(source)

  def move(self):
    return self.pop(0) if len(self) else None

  def current(self):
    return self[0] if len(self) else None


def parse_long(tokens, options):
  """long ::= '--' chars [ ( ' ' | '=' ) chars ] ;"""
  long, eq, value = tokens.move().partition('=')
  assert long.startswith('--')
  value = None if eq == value == '' else value
  similar = [o for o in options if o.long == long]
  if len(similar) > 1:  # might be simply specified ambiguously 2+ times?
    raise DocoptLanguageError('%s is not a unique prefix: %s?' % (long, ', '.join(o.long for o in similar)))
  elif len(similar) < 1:
    argcount = 1 if eq == '=' else 0
    o = Option(None, long, argcount)
    options.append(o)
  else:
    o = Option(similar[0].short, similar[0].long, similar[0].argcount, similar[0].value)
    if o.argcount == 0:
      if value is not None:
        raise DocoptLanguageError('%s must not have an argument' % o.long)
    else:
      if value is None:
        if tokens.current() in [None, '--']:
          raise DocoptLanguageError('%s requires argument' % o.long)
        value = tokens.move()
  return [o]


def parse_shorts(tokens, options):
  """shorts ::= '-' ( chars )* [ [ ' ' ] chars ] ;"""
  token = tokens.move()
  assert token.startswith('-') and not token.startswith('--')
  left = token.lstrip('-')
  parsed = []
  while left != '':
    short, left = '-' + left[0], left[1:]
    similar = [o for o in options if o.short == short]
    if len(similar) > 1:
      raise DocoptLanguageError('%s is specified ambiguously %d times' % (short, len(similar)))
    elif len(similar) < 1:
      o = Option(short, None, 0)
      options.append(o)
    else:  # why copying is necessary here?
      o = Option(short, similar[0].long, similar[0].argcount, similar[0].value)
      value = None
      if o.argcount != 0:
        if left == '':
          if tokens.current() in [None, '--']:
            raise DocoptLanguageError('%s requires argument' % short)
          value = tokens.move()
        else:
          value = left
          left = ''
    parsed.append(o)
  return parsed


def parse_pattern(source, options):
  tokens = Tokens.from_pattern(source)
  result = parse_expr(tokens, options)
  if tokens.current() is not None:
    raise DocoptLanguageError('unexpected ending: %r' % ' '.join(tokens))
  return Required(*result)


def parse_expr(tokens, options):
  """expr ::= seq ( '|' seq )* ;"""
  seq = parse_seq(tokens, options)
  if tokens.current() != '|':
    return seq
  result = [Required(*seq)] if len(seq) > 1 else seq
  while tokens.current() == '|':
    tokens.move()
    seq = parse_seq(tokens, options)
    result += [Required(*seq)] if len(seq) > 1 else seq
  return [Either(*result)] if len(result) > 1 else result


def parse_seq(tokens, options):
  """seq ::= ( atom [ '...' ] )* ;"""
  result = []
  while tokens.current() not in [None, ']', ')', '|']:
    atom = parse_atom(tokens, options)
    if tokens.current() == '...':
      atom = [OneOrMore(*atom)]
      tokens.move()
    result += atom
  return result


def parse_atom(tokens, options):
  """atom ::= '(' expr ')' | '[' expr ']' | 'options'
       | long | shorts | argument | command ;
  """
  token = tokens.current()
  result = []
  if token in '([':
    tokens.move()
    matching, pattern = {'(': [')', Required], '[': [']', Optional]}[token]
    result = pattern(*parse_expr(tokens, options))
    if tokens.move() != matching:
      raise DocoptLanguageError("unmatched '%s'" % token)
    return [result]
  elif token == 'options':
    tokens.move()
    return [OptionsShortcut()]
  elif token.startswith('--') and token != '--':
    return parse_long(tokens, options)
  elif token.startswith('-') and token not in ('-', '--'):
    return parse_shorts(tokens, options)
  elif token.startswith('<') and token.endswith('>') or token.isupper():
    return [Argument(tokens.move())]
  else:
    return [Command(tokens.move())]


def parse_defaults(doc):
  defaults = []
  for s, _ in parse_section('options:', doc):
    # FIXME corner case "bla: options: --foo"
    _, _, s = s.partition(':')  # get rid of "options:"
    split = re.split('\n[ \t]*(-\S+?)', '\n' + s)[1:]
    split = [s1 + s2 for s1, s2 in zip(split[::2], split[1::2])]
    options = [Option.parse(s) for s in split if s.startswith('-')]
    defaults += options
  return defaults


def parse_section(name, source):
  pattern = re.compile('^([^\n]*' + name + '[^\n]*\n?(?:[ \t].*?(?:\n|$))*)', re.IGNORECASE | re.MULTILINE)
  return [(m.group(0).strip(), m) for m in pattern.finditer(source)]


def formal_usage(section):
  _, _, section = section.partition(':')  # drop "usage:"
  pu = section.split()
  return '( ' + ' '.join(') | (' if s == pu[0] else s for s in pu[1:]) + ' )'



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['docopt_sh']

package_data = \
{'': ['*']}

install_requires = \
['docopt', 'termcolor']

extras_require = \
{'test': ['pytest', 'pytest-xdist', 'pytest-pep8']}

entry_points = \
{'console_scripts': ['docopt.sh = docopt_sh.__main__:main']}

setup(name='docopt-sh',
      version='0.9.10',
      description='docopt.sh - Bash argument parser generator.',
      author='Anders Ingemann',
      author_email='anders@ingemann.de',
      url='https://github.com/andsens/docopt.sh',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.5',
     )



#!/usr/bin/env python3
"""
Downloads, extracts, configures, and compiles bash versions for testing
Usage:
  get_bash.py VERSION...
"""

import os
import logging
import urllib.request
import tarfile
import subprocess
import docopt

log = logging.getLogger('get_bash.py')
logging.basicConfig(level=logging.INFO)


def get_bash(version):
  try:
    versions_path = os.path.join(os.path.dirname(__file__), 'bash-versions')
    if not os.path.exists(versions_path):
      os.mkdir(versions_path)
    version_path = os.path.join(versions_path, 'bash-%s' % version)
    executable_path = os.path.join(version_path, 'bash')
    if not os.path.exists(executable_path):
      log.warning('bash-%s executable not found in tests/bash-versions, downloading & compiling now' % version)
      archive_path = os.path.join(versions_path, 'bash-%s.tar.gz' % version)
      if os.path.exists(archive_path):
        log.warning('%s already exists, skipping download' % archive_path)
      else:
        url = 'http://ftp.gnu.org/gnu/bash/bash-%s.tar.gz' % version
        urllib.request.urlretrieve(url, archive_path)
      if os.path.exists(version_path):
        log.warning('%s already exists, skipping extraction' % version_path)
      else:
        with tarfile.open(archive_path) as archive:
          archive.extractall(versions_path)
      makefile_path = os.path.join(version_path, 'Makefile')
      if os.path.exists(makefile_path):
        log.warning('%s already exists, skipping ./configure' % makefile_path)
      else:
        process = subprocess.run('./configure', cwd=version_path)
        process.check_returncode()
      process = subprocess.run('make', cwd=version_path)
      process.check_returncode()
    else:
      log.info('bash-%s is already installed' % version)
  except subprocess.CalledProcessError:
    log.error('Failed to install bash-%s' % version)
    raise

if __name__ == '__main__':
  params = docopt.docopt(__doc__)
  for version in params['VERSION']:
    get_bash(version)



"""
docopt.sh - Bash argument parser generator.
  This program looks for `DOC="... Usage​: ..."`  in a script
  and appends a matching parser to it.
"""
# The "... Usage: ..." above contains a zero-width space between `Usage` and `:`
# in order to prevent docopt from parsing it as a `usage:`` section

__all__ = ['docopt_sh']
__version__ = '0.9.10'


class DocoptError(Exception):
  def __init__(self, message, exit_code=1):
    self.message = message
    self.exit_code = exit_code



import re
from shlex import quote
from collections import OrderedDict
from collections.abc import Iterable
from itertools import chain


class Code(object):

  def __init__(self, code):
    self.code = self._get_iter(code)

  def _get_iter(self, code):
    if isinstance(code, str):
      return iter([code])
    else:
      return iter(code)

  def minify(self, max_line_length):
    return Code(minify(str(self), max_line_length))

  def replace_literal(self, replacements):
    def gen_replace():
      for part in self.code:
        code = str(part)
        for placeholder, replacement in replacements.items():
          code = code.replace(placeholder, str(replacement))
        yield code
    return Code(gen_replace())

  def __iter__(self):
    return self.code

  def __add__(self, other):
    return Code(chain(self.code, self._get_iter(other)))

  def __str__(self):
    return '\n'.join(map(str, self))


def indent(script, level=1):
  indentation = '  ' * level
  return '\n'.join(map(lambda l: indentation + l, script.split('\n')))


def bash_variable_name(name):
  return re.sub(r'^[^a-z_]|[^a-z0-9_]', '_', name, 0, re.IGNORECASE)


def bash_variable_value(value):
  if value is None:
    return ''
  if type(value) is bool:
    return 'true' if value else 'false'
  if type(value) is int:
    return str(value)
  if type(value) is str:
    return quote(value)
  if type(value) is list:
    return '(%s)' % ' '.join(bash_variable_value(v) for v in value)
  raise Exception('Unhandled value type %s' % type(value))


def bash_ifs_value(value):
  if value is None or value == '':
    return "''"
  if type(value) is bool:
    return 'true' if value else 'false'
  if type(value) is int:
    return str(value)
  if type(value) is str:
    return quote(value)
  if type(value) is list:
    raise Exception('Unable to convert list to bash value intended for an IFS separated field')
  raise Exception('Unhandled value type %s' % type(value))


def minify(parser_str, max_length):
  lines = parser_str.split('\n')
  lines = remove_leading_spaces(lines)
  lines = remove_empty_lines(lines)
  lines = remove_newlines(lines, max_length)
  return '\n'.join(lines) + '\n'


def remove_leading_spaces(lines):
  for line in lines:
    yield re.sub(r'^\s*', '', line)


def remove_empty_lines(lines):
  for line in lines:
    if line != '':
      yield line


def remove_newlines(lines, max_length):
  def is_comment(line):
    return re.match(r'^\s*#', line) is not None

  def needs_separator(line):
    return re.search(r'; (then|do)$|else$|\{$', line) is None

  def has_continuation(line):
    return re.search(r'\\\s*$', line) is not None

  def remove_continuation(line):
    return re.sub(r'\s*\\\s*$', '', line)

  def combine(line1, line2):
    if is_comment(line1) or is_comment(line2):
      if not is_comment(line1):
        return line1
      if not is_comment(line2):
        return line2
      return None
    if has_continuation(line1):
      return remove_continuation(line1) + ' ' + line2
    if needs_separator(line1):
      return line1 + '; ' + line2
    else:
      return line1 + ' ' + line2

  previous = next(lines)
  for line in lines:
    combined = combine(previous, line)
    if combined is None:
      previous = next(lines, None)
    elif len(combined) > max_length:
      yield previous
      previous = line
    else:
      previous = combined
  if previous:
    yield previous
